/*
 * Copyright (C) 2023 neocotic
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/* istanbul ignore file */

import { glob } from 'glob';
import { trim } from 'lodash-es';
import mime from 'mime';
import { mkdirp } from 'mkdirp';
import { access, readFile, writeFile } from 'node:fs/promises';
import path from 'node:path';
import { rimraf } from 'rimraf';

/**
 * Contains information for a file.
 *
 * Care should be taken when using this information to generate file paths as the parent directory path and/or name may
 * contain a expression that needs to be evaluated beforehand.
 *
 * @public
 */
export class File {

  /**
   * A convenient shorthand for <code>fs.access</code> wrapped in a <code>Promise</code>.
   *
   * @param {string} filePath - the path of the file to be checked
   * @param {number} [mode] - the mode to be used
   * @return {Promise<void>} A <code>Promise</code> wrapper for <code>fs.access</code>.
   * @public
   */
  static access(filePath, mode) {
    return access(filePath, mode);
  }

  /**
   * A convenient shorthand for <code>rimraf</code> wrapped in a <code>Promise</code>.
   *
   * @param {string} filePath - the path of the file to be deleted
   * @param {Object} [options] - the options to be used
   * @return {Promise<void>} A <code>Promise</code> wrapper for <code>rimraf</code>.
   * @public
   */
  static deleteFile(filePath, options) {
    return rimraf(filePath, options);
  }

  /**
   * Attempts to derive the file format from the specified <code>fileName</code>.
   *
   * Optionally, <code>format</code> can be provided if there's an potential source of the file format already.
   *
   * <code>null</code> will be returned if the format could not be derived.
   *
   * @param {?string} fileName - the name of the file whose format is to be derived (may be <code>null</code>, in which
   * case <code>format</code> is the only chance at deriving the file format)
   * @param {?string} [format] - a potential source of the file format
   * @return {?string} The derived file format or <code>null</code> if it is not possible to be derived based on the
   * information provided.
   * @public
   */
  static deriveFormat(fileName, format) {
    format = format ? trim(format).toLowerCase() : null;

    if (!format && fileName) {
      format = path.extname(fileName).substring(1).toLowerCase() || null;
    }

    return format || null;
  }

  /**
   * A convenient shorthand for <code>glob</code> wrapped in a <code>Promise</code>.
   *
   * @param {string} pattern - the pattern for the files to be found
   * @param {Object} [options] - the options to be used
   * @return {Promise<string[]>} A <code>Promise</code> wrapper for <code>glob</code>.
   * @public
   */
  static findFiles(pattern, options) {
    return glob(pattern, options);
  }

  /**
   * A convenient shorthand for <code>fs.readFile</code> wrapped in a <code>Promise</code>.
   *
   * @param {string} filePath - the path of the file to be read
   * @param {Object|string} [options] - the options to be used
   * @return {Promise<Buffer|string>} A <code>Promise</code> wrapper for <code>fs.readFile</code>.
   * @public
   */
  static readFile(filePath, options) {
    return readFile(filePath, options);
  }

  /**
   * A convenient shorthand for <code>fs.writeFile</code> wrapped in a <code>Promise</code>.
   *
   * This method also ensures that the parent directory for <code>filePath</code> exists before attempting to write to
   * the file.
   *
   * @param {string} filePath - the path of the file to which <code>data</code> is to be written
   * @param {Buffer|string|Uint8Array} data - the data to be written
   * @param {Object|string} [options] - the options to be used
   * @return {Promise<void>} A <code>Promise</code> wrapper for <code>fs.writeFile</code>.
   * @public
   */
  static async writeFile(filePath, data, options) {
    const dirPath = path.dirname(filePath);
    await mkdirp(dirPath);

    return writeFile(filePath, data, options);
  }

  /**
   * @type {?string}
   * @private
   */
  #dir;
  /**
   * @type {?string}
   * @private
   */
  #name;
  /**
   * @type {?string}
   * @private
   */
  #format;
  /**
   * @type {Config}
   * @private
   */
  #config;
  /**
   * @type {boolean}
   * @private
   */
  #evaluated;

  /**
   * Creates an instance of {@link File}.
   *
   * Optionally, <code>evaluated</code> should be set to <code>true</code> if <code>dir</code> and <code>name</code>
   * have already been evaluated (e.g. via {@link Config#evaluate}) to avoid double-evaluation.
   *
   * @param {?string} dir - the parent directory path to be used (may be <code>null</code>)
   * @param {?string} name - the name to be used (may be <code>null</code>)
   * @param {?string} format - the format to be used (may be <code>null</code>)
   * @param {Config} config - the {@link Config} to be used
   * @param {boolean} [evaluated] - <code>true</code> if <code>dir</code> and <code>name</code> have already been
   * evaluated; otherwise <code>false</code>
   * @public
   */
  constructor(dir, name, format, config, evaluated = false) {
    this.#dir = dir;
    this.#name = name;
    this.#format = format;
    this.#config = config;
    this.#evaluated = evaluated;
  }

  /**
   * Returns the base name for this {@link File}.
   *
   * Optionally, <code>excludeExtension</code> can be used to exclude the extension from the base name. The extension
   * will be derived from the name, where possible, falling back to the format, where available.
   *
   * <code>null</code> will be returned if the name of this {@link File} is unavailable.
   *
   * @param {boolean} [excludeExtension] - <code>true</code> to exclude the extension from the base name; otherwise
   * <code>false</code>.
   * @return {?string} The base name, potentially with the extension excluded, or <code>null</code> if it has no name.
   * @public
   */
  base(excludeExtension = false) {
    const name = this.#name;
    if (!name) {
      return null;
    }

    return excludeExtension ? path.basename(name, this.extension() || '') : path.basename(name);
  }

  /**
   * Creates an instance of {@link File} based on the directory, name, and format of this {@link File} while using the
   * default values provided for those fields when unavailable.
   *
   * Optionally, <code>evaluated</code> should be set to <code>true</code> if <code>defaultDir</code> and
   * <code>defaultName</code>, or the corresponding fields on this {@link File}, have already been evaluated (e.g. via
   * {@link Config#evaluate}) to avoid double-evaluation.
   *
   * @param {string} defaultDir - the default directory path to be used
   * @param {string} defaultName - the default name to be used
   * @param {string} defaultFormat - the default format to be used
   * @param {boolean} [evaluated] - <code>true</code> if <code>defaultDir</code> and <code>defaultName</code> (or the
   * corresponding fields on this {@link File}) have already been evaluated; otherwise <code>false</code>
   * @return {File} A {@link File} based on this instance that potentially used the default values.
   * @public
   */
  defaults(defaultDir, defaultName, defaultFormat, evaluated = false) {
    const dir = this.#dir || defaultDir;
    const name = this.#name || defaultName;
    const format = this.#format || defaultFormat;

    return new File(dir, name, format, this.#config, evaluated);
  }

  /**
   * Returns an instance of {@link File} based on this {@link File}, whose diretory and name have been evaluated, by
   * interpolating data properties and executing embedded JavaScript within those fields.
   *
   * Optionally, <code>additionalData</code> can be provided to expose more variables to <code>expression</code> during
   * evaluation.
   *
   * A <code>config</code> variable can be used within the expression to reference the {@link Config} for this
   * {@link File}.
   *
   * A reference to this {@link File} will be returned if it has previously been evluated in an attempt to prevent
   * double-evaluation.
   *
   * @param {Object} [additionalData] - an object whose properties will be exposed as variables when the expressions are
   * evaluated
   * @return {File} A {@link File} based on this instance but with the directory and name fields evaluated, or exactly
   * this instance, if previously evaluated.
   * @public
   */
  evaluate(additionalData) {
    if (this.#evaluated) {
      return this;
    }

    const { config, dir, name } = this;
    const evaluatedDir = dir ? config.evaluate(dir, additionalData) : null;
    const evaluatedName = name ? config.evaluate(name, additionalData) : null;

    return new File(evaluatedDir, evaluatedName, this.#format, config, true);
  }

  /**
   * Returns the extension for this {@link File}.
   *
   * The extension will be derived from the name, where possible, falling back to the format, where available.
   *
   * <code>null</code> will be returned if the name of this {@link File} is unavailable and/or the extension could not
   * be derived from the name and the format of this {@link File} is unavailable.
   *
   * @return {?string} The extension or <code>null</code> if it has no name or format and/or the extension could not be
   * derived from the name.
   * @public
   */
  extension() {
    const { format, name } = this;
    if (!name) {
      return format ? `.${format}` : null;
    }

    return path.extname(name) || (format && `.${format}`) || null;
  }

  /**
   * Resolves the specified sequence of <code>paths</code> or path segments into an absolute path relative to the parent
   * directory for this {@link File}.
   *
   * @param {...string} paths - the sequence of paths or path segments to be resolved
   * @return {string} An absolute file path.
   * @public
   */
  resolve(...paths) {
    return path.resolve(this.#dir, ...paths);
  }

  /**
   * @override
   */
  toString() {
    return this.absolute;
  }

  /**
   * Returns the absolute path of this {@link File}.
   *
   * @return {string} The absolute file path.
   * @public
   */
  get absolute() {
    return path.join(this.#dir, this.#name);
  }

  /**
   * Returns the {@link Config} for this {@link File}.
   *
   * @return {Config} The configuration.
   * @public
   */
  get config() {
    return this.#config;
  }

  /**
   * Returns the parent directory path of this {@link File}.
   *
   * The directory may be <code>null</code> if it is unavailable.
   *
   * @return {?string} The directory path or <code>null</code> if unavailable.
   * @public
   */
  get dir() {
    return this.#dir;
  }

  /**
   * Returns the format of this {@link File}.
   *
   * The format may be <code>null</code> if it is unavailable.
   *
   * @return {?string} The format or <code>null</code> if unavailable.
   * @public
   */
  get format() {
    return this.#format;
  }

  /**
   * Returns the name of this {@link File}.
   *
   * The name may be <code>null</code> if it is unavailable.
   *
   * @return {?string} The name or <code>null</code> if unavailable.
   * @public
   */
  get name() {
    return this.#name;
  }

  /**
   * Returns the path of this {@link File} relative to the base directory from where the data for the associated
   * {@link Config} was originally loaded.
   *
   * @return {string} The relative file path.
   * @public
   */
  get relative() {
    return this.#config.relative(this.absolute);
  }

  /**
   * Returns the MIME type for this {@link File}.
   *
   * <code>null</code> will be returned if the name of this {@link File} is unavailable and/or the extension could not
   * be derived from the name and the format of this {@link File} is unavailable <b>or</b> a valid MIME could not be
   * derived from that extension.
   *
   * @return {?string} The MIME type or <code>null</code> if it has no name or format and/or the extension could not be
   * derived from the name or the MIME type could not be derived from either that extension.
   * @public
   */
  get type() {
    return mime.getType(this.extension());
  }

}
